thelper.nn.mobilenet — thelper 0.3.8 documentation

### Navigation

- ☰
- thelper-0.3.8 »
- Module code »

Fork me on GitHub

# Source code for thelper.nn.mobilenet

    # MobileNet v2 derived from https://github.com/tonylins/pytorch-mobilenet-v2
    import math

    import torch.nn as nn

    import thelper

    def conv_bn(inp, oup, stride):
        return nn.Sequential(
            nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
            nn.BatchNorm2d(oup),
            nn.ReLU6(inplace=True)
        )

    def conv_1x1_bn(inp, oup):
        return nn.Sequential(
            nn.Conv2d(inp, oup, 1, 1, 0, bias=False),
            nn.BatchNorm2d(oup),
            nn.ReLU6(inplace=True)
        )

    class InvertedResidual(nn.Module):

        def __init__(self, inp, oup, stride, expand_ratio):
            super().__init__()
            self.stride = stride
            if stride not in [1, 2]:
                raise AssertionError("stride should be 1 or 2")
            self.use_res_connect = self.stride == 1 and inp == oup
            self.conv = nn.Sequential(
                # pw
                nn.Conv2d(inp, inp * expand_ratio, 1, 1, 0, bias=False),
                nn.BatchNorm2d(inp * expand_ratio),
                nn.ReLU6(inplace=True),
                # dw
                nn.Conv2d(inp * expand_ratio, inp * expand_ratio, 3, stride, 1, groups=inp * expand_ratio, bias=False),
                nn.BatchNorm2d(inp * expand_ratio),
                nn.ReLU6(inplace=True),
                # pw-linear
                nn.Conv2d(inp * expand_ratio, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )

        def forward(self, x):
            if self.use_res_connect:
                return x + self.conv(x)
            else:
                return self.conv(x)

    class MobileNetV2(thelper.nn.Module):

        def __init__(self, task, input_size=224, width_mult=1.):
            # note: must always forward args to base class to keep backup
            super().__init__(task, input_size=input_size, width_mult=width_mult)
            # setting of inverted residual blocks
            self.interverted_residual_setting = [
                # t, c, n, s
                [1, 16, 1, 1],
                [6, 24, 2, 2],
                [6, 32, 3, 2],
                [6, 64, 4, 2],
                [6, 96, 3, 1],
                [6, 160, 3, 2],
                [6, 320, 1, 1],
            ]
            # building first layer
            if input_size % 32 != 0:
                raise AssertionError("input size should be multiple of 32 px")
            input_channel = int(32 * width_mult)
            self.last_channel = int(1280 * width_mult) if width_mult > 1.0 else 1280
            self.features = [conv_bn(3, input_channel, 2)]
            # building inverted residual blocks
            for t, c, n, s in self.interverted_residual_setting:
                output_channel = int(c * width_mult)
                for i in range(n):
                    if i == 0:
                        self.features.append(InvertedResidual(input_channel, output_channel, s, t))
                    else:
                        self.features.append(InvertedResidual(input_channel, output_channel, 1, t))
                    input_channel = output_channel
            # building last several layers
            self.features.append(conv_1x1_bn(input_channel, self.last_channel))
            self.features.append(nn.AvgPool2d(input_size // 32))
            # make it nn.Sequential
            self.features = nn.Sequential(*self.features)
            # building classifier
            self.classif_features = 1000  # default count, will be updated if needed
            self.classifier = nn.Sequential(
                nn.Dropout(),
                nn.Linear(self.last_channel, self.classif_features),
            )
            self._initialize_weights()
            self.set_task(task)

        def forward(self, x):
            x = self.features(x)
            x = x.view(-1, self.last_channel)
            x = self.classifier(x)
            return x

        def set_task(self, task):
            assert isinstance(task, thelper.tasks.Classification), "missing impl for non-classif task type"
            num_classes = len(task.class_names)
            if num_classes != self.classif_features:
                self.classifier = nn.Sequential(
                    nn.Dropout(),
                    nn.Linear(self.last_channel, num_classes),
                )
                self.classif_features = num_classes
            self.task = task

        def _initialize_weights(self):
            for m in self.modules():
                if isinstance(m, nn.Conv2d):
                    n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                    m.weight.data.normal_(0, math.sqrt(2. / n))
                    if m.bias is not None:
                        m.bias.data.zero_()
                elif isinstance(m, nn.BatchNorm2d):
                    m.weight.data.fill_(1)
                    m.bias.data.zero_()
                elif isinstance(m, nn.Linear):
                    n = m.weight.size(1)
                    m.weight.data.normal_(0, 0.01)
                    m.bias.data.zero_()

### Navigation

- index
- modules |